@@ -149,13 +149,25 @@ static int dm_create_device(int reload, struct crypt_options *options,
                             const char *key)
 {
 	struct dm_task *dmt = NULL;
+	struct dm_task *dmt_query = NULL;
 	struct dm_info dmi;
 	char *params = NULL;
 	int r = -EINVAL;
 
 	params = get_params(options, key);
 	if (!params)
-		goto out;
+		goto out_no_removal;
+
+	/* If the device exists, refuse to run */
+	if (!(dmt_query = dm_task_create(DM_DEVICE_STATUS)))
+		goto out_no_removal;
+	if (!dm_task_set_name(dmt_query, options->name))
+		goto out_no_removal;
+	if (dm_task_run(dmt_query)) {
+		set_error("Volume already open.");
+	        goto out_no_removal;
+	}
+
 	if (!(dmt = dm_task_create(reload ? DM_DEVICE_RELOAD
 	                                  : DM_DEVICE_CREATE)))
 		goto out;
@@ -205,10 +217,14 @@ out_restore_error:
 		if (error)
 			free(error);
 	}
+
+out_no_removal:
 	if (params)
 		safe_free(params);
 	if (dmt)
 		dm_task_destroy(dmt);
+	if(dmt_query)
+		dm_task_destroy(dmt_query);
 	flush_dm_workqueue();
 	return r;
 }
@@ -236,6 +252,11 @@ static int dm_query_device(int details, struct crypt_options *options,
 	if (!dm_task_get_info(dmt, &dmi))
 		goto out;
 
+	if (!dmi.exists) {
+		r = -ENODEV;
+		goto out;
+	}
+
 	next = dm_get_next_target(dmt, next, &start, &length,
 	                          &target_type, &params);
 	if (!target_type || strcmp(target_type, CRYPT_TARGET) != 0 ||
