@@ -8,6 +8,7 @@
 #include <fcntl.h>
 #include <unistd.h>
 #include <errno.h>
+#include <signal.h>
 #include <assert.h>
 
 #include "libcryptsetup.h"
@@ -23,6 +24,12 @@ struct device_infos {
 static int memory_unsafe = 0;
 static char *default_backend = NULL;
 
+static void catch_alarm(int sig_num)
+{
+       fprintf(stderr, "Operation timed out. Exiting.\n");
+       exit(0);
+}
+
 static int setup_enter(struct setup_backend *backend)
 {
 	int r;
@@ -68,13 +75,13 @@ static int setup_leave(struct setup_backend *backend)
 /*
  * Password reading behaviour matrix of get_key
  * 
- *                    p   v   n
- * -----------------+---+---+----
- * interactive      | Y | Y | Y
- * from fd          | N | N | Y
- * from binary file | N | N | N
+ *                    p   v   n   h
+ * -----------------+---+---+---+---
+ * interactive      | Y | Y | Y | Inf
+ * from fd          | N | N | Y | Inf
+ * from binary file | N | N | N | Inf or options->key_size
  *
- * Legend: p..prompt, v..can verify, n..newline-stop
+ * Legend: p..prompt, v..can verify, n..newline-stop, h..read horizon
  * 
  */
 
@@ -82,8 +89,10 @@ static void get_key(struct crypt_options *options, char *prompt, char **key, int
 {
 	int fd;
 	const int verify = options->flags & CRYPT_FLAG_VERIFY;
+	const int verify_if_possible = options->flags & CRYPT_FLAG_VERIFY_IF_POSSIBLE;
 	char *pass = NULL;
 	int newline_stop;
+	int read_horizon; 
 	
 	if(options->key_file) {
 		fd = open(options->key_file, O_RDONLY);
@@ -94,10 +103,22 @@ static void get_key(struct crypt_options *options, char *prompt, char **key, int
 			goto out_err;
 		}	
 		newline_stop = 0;
+
+		/* This can either be 0 (LUKS) or the actually number 
+		   of key bytes (default or passed by -s) */
+		read_horizon = options->key_size;
 	} else {
 		fd = options->passphrase_fd;
 		newline_stop = 1;
+		read_horizon = 0;   /* Infinite, if read from terminal or fd */
 	}	
+
+	signal(SIGALRM, catch_alarm);
+	if(options->timeout) {
+		alarm(options->timeout);
+	} else {
+		alarm(0);
+	}
 	
 	/* Interactive case */
 	if(isatty(fd)) {
@@ -111,7 +132,7 @@ static void get_key(struct crypt_options *options, char *prompt, char **key, int
 		pass = safe_strdup(pass2);
 		memset(pass2, 0, strlen(pass2));
 		
-		if (verify) {
+		if (verify || verify_if_possible) {
 			char *pass_verify = getpass("Verify passphrase: ");
 			if (!pass_verify || strcmp(pass, pass_verify) != 0) {
 				set_error("Passphrases do not match");
@@ -139,7 +160,7 @@ static void get_key(struct crypt_options *options, char *prompt, char **key, int
 		 * such as /dev/random, because in this case, the loop
 		 * will read forever.
 		 */ 
-		if(options->key_file && options->key_size == 0) {
+		if(options->key_file && read_horizon == 0) {
 			struct stat st;
 			if(stat(options->key_file, &st) < 0) {
 		 		set_error("Can't stat key file");
@@ -152,7 +173,7 @@ static void get_key(struct crypt_options *options, char *prompt, char **key, int
 			}
 		}
 		buflen = 0;
-		for(i = 0; options->key_size == 0 || i < options->key_size; i++) {
+		for(i = 0; read_horizon == 0 || i < read_horizon; i++) {
 			if(i >= buflen - 1) {
 				buflen += 128;
 				pass = safe_realloc(pass, buflen);
@@ -170,6 +191,8 @@ static void get_key(struct crypt_options *options, char *prompt, char **key, int
 		pass[i] = 0;
 		*key = pass;
 		*passLen = i;
+
+		alarm(0);
 	}
 	return;
 
@@ -283,14 +306,14 @@ static int parse_into_name_and_mode(const char *nameAndMode, char *name,
 	// Token content stringification, see info cpp/stringification
 #define str(s) #s
 #define xstr(s) str(s)
-#define sp1 "%" xstr(LUKS_CIPHERNAME_L) "[^-]-%" xstr(LUKS_CIPHERMODE_L)  "s"
-#define sp2 "%" xstr(LUKS_CIPHERNAME_L) "[^-]"
+#define scanpattern1 "%" xstr(LUKS_CIPHERNAME_L) "[^-]-%" xstr(LUKS_CIPHERMODE_L)  "s"
+#define scanpattern2 "%" xstr(LUKS_CIPHERNAME_L) "[^-]"
 
 	int r;
 
-	if(sscanf(nameAndMode,sp1, name, mode) != 2) {
-		if((r = sscanf(nameAndMode,sp2,name)) == 1) {
-			strncpy(mode,"cbc-plain",10);
+	if(sscanf(nameAndMode,scanpattern1, name, mode) != 2) {
+		if((r = sscanf(nameAndMode,scanpattern2,name)) == 1) {
+			strncpy(mode,"cbc-essiv:sha256",10);
 		} 
 		else {
 			fprintf(stderr, "no known cipher-spec pattern detected\n");
@@ -556,9 +579,8 @@ static int __crypt_luks_open(int arg, struct setup_backend *backend, struct cryp
 		r = -EINVAL; goto out2;
 	}
 	options->size -= options->offset;
-	backend->create(0, options, mk.key);
+	r = backend->create(0, options, mk.key);
 
-	r = 0;
  out2:
 	free(dmCipherSpec);
  out1:
